class BatchItemNotRunError(Exception):
    pass


class BatchItemUnsuccessfulError(Exception):
    pass


class WrongAlgorithmExtensionError(Exception):
    pass


class DependencyError(Exception):
    pass
